import functools
import os
import ssl
import uuid
from typing import Any, Dict, List, Literal, Optional, Tuple

import jinja2
import pydantic
import yaml

from .exceptions import TplBuildException, TplBuildTemplateException


def _normalize_rel_path(path: str) -> str:
    """Normalize and coerce a path into a relative path."""
    return f".{os.path.sep}{os.path.normpath(os.path.join(os.path.sep, path))[1:]}"


class BaseModel(pydantic.BaseModel):
    """Base model class configuration"""

    class Config:
        """Default config for our models"""

        extra = pydantic.Extra.forbid


class TplContextConfig(BaseModel):
    """
    Config model representing a build context.
    """

    #: The base directory (relative to the config base directory) of
    #:     the build context. This must be a relative path and cannot point
    #:     above the config base directory.
    base_dir: str = "."
    #: The umask as a three digit octal string. This may also be set to
    #:     None if the context permissions should be passed through directly.
    #:     This parameter helps ensure base images are stable across developers
    #:     who may have different group/all permissions set.
    umask: Optional[str] = "022"
    #: The ignore_file to load patterns from. If this and :attr:`ignore`
    #:     are both None then this will attempt to load ".dockerignore", using
    #:     an empty list of patterns if that cannot be loaded.
    ignore_file: Optional[str] = None
    #: Ignore file string. If present this will be used over :attr:`ignore_file`.
    ignore: Optional[str] = None

    @pydantic.validator("umask")
    def umask_valid_octal(cls, v):
        """Ensure that umask is three-digit octal sequence"""
        if v is None:
            return v
        if 0 <= int(v, 8) <= 0o777:
            raise ValueError("umask out of range")
        return v

    @pydantic.validator("base_dir")
    def normalize_base_dir(cls, v):
        """Normalize the base directory"""
        return _normalize_rel_path(v)


class ClientCommand(BaseModel):
    """Configuration to invoke an external build command."""

    #: A jinja template used to construct invoke arguments and environment
    #: variables based on the template arguments passed. Depending on the
    #: build command different template arguments may be passed. All templates
    #: are passed an `args` list and an `environment` dict that they should
    #: populate with the command arguments and environment variables used
    #: to invoke the build command. The output of the template will be ignored.
    template: str

    def render(
        self,
        jinja_env: jinja2.Environment,
        params: Dict[str, str],
    ) -> Tuple[List[str], Dict[str, str]]:
        """
        Return the list of arguments and addition environment variables to
        pass to execute the rendered command.
        """
        args: List[str] = []
        environment: Dict[str, str] = {}

        try:
            for _ in jinja_env.from_string(self.template).generate(
                **params,
                args=args,
                environment=environment,
            ):
                pass
        except jinja2.TemplateError as exc:
            raise TplBuildTemplateException(
                "Failed to render command template"
            ) from exc
        if not args:
            print(self.template)
            raise TplBuildException("command template rendered no command arguments")
        return args, environment


class ClientConfig(BaseModel):
    """
    Configuration of commands to perform various container operations. This is
    meant to be a generic interface that could plug into a variety of container
    build systems. Typically you can just set :attr:`UserConfig.client_type` to
    select from preconfigured client configurations.
    """

    #: Build command config template. This should render an appropriate command
    #: to build an image using a dockerfile named "Dockerfile" and build
    #: context provided by stdin. The output should be tagged as the passed
    #: argument `image`.
    #:
    #: Arguments:
    #:   image: str - The image name to tag the output
    #:   platform: str? - The build platform to use if known.
    build: ClientCommand
    #: Tag command config template. This should tag an existing image with
    #: a new image name.
    #:
    #: Arguments:
    #:   source_image: str - The source image name
    #:   dest_image: str - The new name to tag `source_image` as
    tag: ClientCommand
    #: Pull command config template. This should pull the named image from
    #: the remote registry into local storage.
    #:
    #: Arguments:
    #:   image: str - The name of the image to pull
    pull: Optional[ClientCommand] = None
    #: Push command config template. This should push the named image to
    #: the remote registry from local storage.
    #:
    #: Arguments:
    #:   image: str - The name of the image to push
    push: ClientCommand
    #: Un-tag command config template. This should untag the named image
    #: allowing data referenced by the image to be reclaimed.
    #:
    #: Arguments:
    #:   image: str - The name of the image to untag
    untag: ClientCommand
    #: Command that should print out the default build platform for the client.
    #: This template is passed no additional arguments. If this command is not
    #: available the default build platform will be calculated using the local
    #: client platform instead. The output will be normalized to convert
    #: e.g. "linux/x64_64" to "linux/amd64". This will only be used for
    #: platform aware build configurations.
    platform: Optional[ClientCommand] = None


UNSET_CLIENT_CONFIG = ClientConfig(
    build=ClientCommand(template=""),
    tag=ClientCommand(template=""),
    push=ClientCommand(template=""),
    untag=ClientCommand(template=""),
)


@functools.lru_cache
def get_builtin_configs() -> Dict[str, ClientConfig]:
    """
    Return a cached mapping of preconfigured clients.
    """
    path = os.path.join(os.path.dirname(__file__), "builtin_clients.yml")
    with open(path, "r", encoding="utf-8") as fdata:
        configs = yaml.safe_load(fdata)
    return {
        config_name: ClientConfig(**config_data)
        for config_name, config_data in configs.items()
    }


class UserSSLContext(BaseModel):
    """Custom SSL context used to contact registries"""

    #: Disable SSL/TLS verification
    insecure: bool = False
    #: File path to load CA certificates to trust.
    cafile: Optional[str] = None
    #: Folder container CA certificate files to trust.
    capath: Optional[str] = None
    #: Raw certificate data to trust.
    cadata: Optional[str] = None
    #: If True default system certs will be loaded in addition to any certs
    #: implied by `cafile`, `capath`, or `cadata`. Normally these will only be
    #: loaded if those are all unset.
    load_default_certs: bool = False

    def create_context(self) -> ssl.SSLContext:
        """Returns a SSLContext constructed from the passed options"""
        ctx = ssl.create_default_context(
            cafile=self.cafile,
            capath=self.capath,
            cadata=self.cadata,
        )
        if self.insecure:
            ctx.check_hostname = False
            ctx.verify_mode = ssl.CERT_NONE
        if self.load_default_certs:
            ctx.load_default_certs()
        return ctx


class StageConfig(BaseModel):
    """Configuration data for a named build stage"""

    #: Is the stage a base stage. If True then :attr:`image_names` and
    #: :attr:`push_names` must be left as None.
    base: bool = False
    #: All image names to assign to the built image. These values are templates
    #: that will be rendered in the same manner as
    #: :attr:`tplbuild.config.TplConfig.stage_image_name`. If left as None
    #: TplConfig.stage_image_name will be used instead as the default.
    image_names: Optional[List[str]] = None
    #: All image names to assign and then push to remote registries when
    #: when publishing images. These values are templates that will be rendered
    #: in the same manner as
    #: :attr:`tplbuild.config.TplConfig.stage_push_name`. If left as None
    #: TplConfig.stage_push_name will be used instead as the default.
    push_names: Optional[List[str]] = None

    @pydantic.validator("image_names")
    def image_names_empty_for_base(cls, v, values):
        """Ensure base images have no image_names"""
        if v is not None and values["base"]:
            raise ValueError("image_names must be empty for base images")
        return v

    @pydantic.validator("push_names")
    def push_names_empty_for_base(cls, v, values):
        """Ensure base images have no push_names"""
        if v is not None and values["base"]:
            raise ValueError("push_names must be empty for base images")
        return v


class UserConfig(BaseModel):
    """User settings controlling tplbuild behavior"""

    #: Must be "1.0"
    version: Literal["1.0"] = "1.0"
    #: If :attr:`client` is None this field will be used to set the client
    #: configuration. Supported values are currently "docker" and "podman".
    #: If :attr:`client` is not None this field is ignored.
    client_type: str = "docker"
    #: Client commands to use to perform different container actions. If unset
    #: a default configuration will be provided based on the value of
    #: :attr:`client_type`. If you wish to use a different builder or supply
    #: additional arguments to the build this would be the place to do it.
    client: ClientConfig = UNSET_CLIENT_CONFIG
    #: Maximum number of concurrent build jbs. If set to 0 this will be set to
    #: `os.cpu_count()`.
    build_jobs: int = 0
    #: Maximum number of concurrent push or pull jobs.
    push_jobs: int = 4
    #: Maximum number of concurrent tag jobs.
    tag_jobs: int = 8
    #: Maximum number of times a build will be retried before failing a build.
    build_retry: int = 0
    #: Maximum number of times a push or pull will be retried before failing a build.
    push_retry: int = 0
    #: Configure the SSL context used to contact registries. This only controls
    #: accesses made by tplbuild itself. The client builder may need to be
    #: configured separately.
    ssl_context: UserSSLContext = UserSSLContext()
    #: The path to the container auth configuration file to use when contacting
    #: registries. By default this will check the default search paths and conform
    #: to the syntax described in
    #: https://github.com/containers/image/blob/main/docs/containers-auth.json.5.md.
    auth_file: Optional[str] = None

    @pydantic.validator("build_jobs", always=True)
    def build_jobs_valid(cls, v):
        """ensure build_jobs is non-negative"""
        if v == 0:
            return os.cpu_count() or 4
        if v < 0:
            raise ValueError("build_jobs must be non-negative")
        return v

    @pydantic.validator("push_jobs")
    def push_jobs_valid(cls, v):
        """ensure push_jobs is positive"""
        if v <= 0:
            raise ValueError("push_jobs must be positive")
        return v

    @pydantic.validator("tag_jobs")
    def tag_jobs_valid(cls, v):
        """ensure tag_jobs is positive"""
        if v <= 0:
            raise ValueError("push_jobs must be positive")
        return v

    @pydantic.validator("client", always=True)
    def default_replace_client(cls, v, values):
        """replace client with client_type if unset"""
        if v.build.template:
            return v
        client_type = values["client_type"]
        v = get_builtin_configs().get(client_type)
        if v is None:
            raise ValueError(f"no builtin client type named {repr(client_type)}")
        return v


class TplConfig(BaseModel):
    """Configuration settings for a single tplbuild project"""

    #: Must be "1.0"
    version: Literal["1.0"] = "1.0"
    #: Jinja template that renders to the image repo where a base image
    #: will be stored. This should *not* include a tag as tplbuild uses
    #: the tag itself to identify the content-addressed build. This
    #: template is passed "stage_name", "profile", and "platform"
    #: corresponding to the name of the stage, the name of the profile
    #: that rendered the image, and the name of the build platform respectively.
    base_image_repo: Optional[str] = None
    #: A Jinja template that renders to the default image name for a
    #: given stage_name. This template is passed "vars",
    #: "stage_name", "profile", "and "platform" parameters.
    stage_image_name: str = "{{ stage_name}}"
    #: A Jinja template that renders to the default push name for a
    #: given stage_name. This template is passed "vars",
    #: "stage_name", "profile", "and "platform" parameters.
    stage_push_name: str = "{{ stage_name}}"
    #: The dockerfile "syntax" to use as the build frontend when running against
    #: builders that understand the "syntax" directive. For some build clients
    #: specifiying a syntax may be required (e.g. the buildx client requires
    #: docker/dockerfile:1.4 or later).
    dockerfile_syntax: pydantic.constr(regex=r"^[^\s]*$") = ""  # type: ignore
    #: List of platforms to build images for. This defaults to linux/amd64
    #:     but should be explicitly configured. Base images will be built
    #:     for each platform listed here allowing for top-level images to
    #:     be built in any of those platforms or as manifest lists when
    #:     pushed.
    platforms: List[str] = ["linux/amd64"]
    #: A mapping of profile names to string-key template arguments to pass
    #:     to any documents rendered through Jinja for each profile. Defaults
    #:     to a single empty profile named 'default'.
    profiles: Dict[str, Dict[str, Any]] = {"default": {}}
    #: The name of the default profile to use. If this is empty
    #:     the first profile name from :attr:`profiles` will be used instead.
    #:     Profile data will be made availble when rendering under the
    #:     variable name `vars`.
    default_profile: str = ""
    #: A set of named build context configurations. These contexts may
    #:     be referred to by name in the build file and should be unique
    #:     among all other stages. If there is a context named "default" it
    #:     will be used as the default build context. Otherwise the first
    #:     listed context will be treated as the default. Any COPY instruction
    # ;     may use --from=context_name to copy from a specific named context.
    contexts: Dict[str, TplContextConfig] = {"default": TplContextConfig()}
    #: A mapping of stage names to stage configs. This can be used to override
    #: the default behavior of tplbuild or apply different or more than just a
    #: single image name to a given stage. See
    #: :meth:`Tplbuild.default_stage_config` for information about default stage
    #: configuration.
    stages: Dict[str, StageConfig] = {}
    #: Search directories for included tempates. Paths must be relative to the
    #: project base directory.
    template_paths: List[str] = ["."]
    #: Template entrypoint to render to generate all build stages. Path should
    #: be relative to one of the `template_paths`.
    template_entrypoint: str = "Dockerfile"

    @pydantic.validator("platforms")
    def platform_nonempty(cls, v):
        """Ensure that platforms is non empty"""
        if not v:
            raise ValueError("platforms cannot be empty")
        return v

    @pydantic.validator("profiles")
    def profile_not_empty(cls, v):
        """Make sure there is at least one profile"""
        if not v:
            raise ValueError("profiles cannot be empty")
        return v

    @pydantic.validator("profiles")
    def profile_name_nonempty(cls, v):
        """Make sure all profile names are non-empty"""
        if any(profile_name == "" for profile_name in v):
            raise ValueError("profile name cannot be empty")
        return v

    @pydantic.validator("default_profile")
    def default_profile_exists(cls, v, values):
        """Make sure default profile name exists if non-empty"""
        if v and v not in values["profiles"]:
            raise ValueError("default_profile must be a valid profile name")
        return v

    @pydantic.validator("template_paths")
    def normalize_template_paths(cls, v):
        """Normalize the template search paths"""
        return [_normalize_rel_path(path) for path in v]

    @pydantic.validator("template_entrypoint")
    def normalize_template_entrypoint(cls, v):
        """Normalize the template entrypoint path"""
        return v  # _normalize_rel_path(v)


class BaseImageBuildData(BaseModel):
    """
    Stores content addressed keys used to store and retrieve base images
    from the remote registry.
    """

    #: The hash of all inputs that go into defining the image build definition.
    build_hash: str
    #: The image digest stored in the registry.
    image_digest: str


class BuildData(BaseModel):
    """
    Any build data that is managed by tplbuild itself rather than being
    configuration data provided by the user. Right now this includes a
    mapping of source images and base images to their content address
    sources.
    """

    #: Must be "1.0"
    version: Literal["1.0"] = "1.0"
    #: Mapping of repo -> tag -> platform -> source image manifest digest.
    source: Dict[str, Dict[str, Dict[str, str]]] = {}
    #: Mapping of profile -> stage_name -> platform -> cached base image
    #: build data.
    base: Dict[str, Dict[str, Dict[str, BaseImageBuildData]]] = {}
    #: A string combined with the base image definition hashes to produce
    #: the final hash for base images. This ensures that different projects
    #: use disjoint hash spaces, that the base image keys bear no information
    #: by themselves, and to force rebuilds by changing the salt.
    hash_salt: str = ""

    @pydantic.validator("hash_salt", always=True)
    def default_profile_exists(cls, v):
        """Fill in hash salt if not set"""
        if not v:
            return str(uuid.uuid4())
        return v



from typing import List

from tplbuild.plan import BuildOperation
from tplbuild.tplbuild import TplBuild


def debug_build_operations(tplbld: TplBuild, build_ops: List[BuildOperation]) -> None:
    """
    Print out rendered build operations for debugging purposes.
    """
    first = True
    rendered_ops = tplbld.executor.render_build_ops(build_ops)
    for rendered_op in rendered_ops:
        if not first:
            print("")
        first = False

        print(f"# Building {rendered_op.build_title}")
        print(rendered_op.dockerfile)

        for tag, push_tag in rendered_op.tags.items():
            if push_tag:
                print(f"# Push {rendered_op.build_title} as {tag}")
            else:
                print(f"# Tag {rendered_op.build_title} as {tag}")



import abc
from dataclasses import dataclass
from typing import Any, Dict, Iterable, List, Optional, Set

from .config import StageConfig
from .context import BuildContext
from .exceptions import TplBuildException
from .utils import extract_command_flags, split_line_tokens


class ImageDefinition(metaclass=abc.ABCMeta):
    """
    Base class for all image deinitions. These abstractly represent a build
    graph.
    """

    @abc.abstractmethod
    def local_hash_data(self, symbolic: bool) -> Any:
        """
        Return a JSON-able payload suitable for hashing the image node. This
        should not recursively include hashes of any dependencies.
        """
        return None

    def get_dependencies(self) -> List["ImageDefinition"]:
        """
        Returns a list of image dependencies for this image node. The first
        dependant is considered the "primary" dependant.
        """
        return []

    def set_dependencies(self, deps: Iterable["ImageDefinition"]) -> None:
        """
        Sets the dependencies for this image. This must be the correct size
        for the image type.
        """
        assert not tuple(deps)

    def merge_into(self, image: "ImageDefinition") -> None:
        """
        Called during planning during the canonicalization phase. Two images will
        be merged if they represent the same build step. This step allows for
        bookkeepping when tracking what image definitions came from where.
        """


@dataclass(frozen=True)
class StageDescriptor:
    """Describes a rendered stage"""

    name: str
    profile: str
    platform: str


@dataclass(eq=False)  # type: ignore
class StageDefinedImage(ImageDefinition):  # pylint: disable=abstract-method
    """
    Parent class for image nodes that are defined by build stages. This is
    meant to help facilitate bookkeeping for display/debug purposes of where
    an image node was defined.
    """

    stage_descs: Set[StageDescriptor]

    def merge_into(self, image: ImageDefinition) -> None:
        """Merge the stage descriptors together."""
        assert isinstance(image, StageDefinedImage)
        self.stage_descs.update(image.stage_descs)


@dataclass(eq=False)
class CommandImage(StageDefinedImage):
    """Image node ending in a command other than COPY."""

    parent: ImageDefinition
    command: str
    args: str

    def local_hash_data(self, symbolic: bool) -> Any:
        """Return the local hash data for this node."""
        return [
            self.command,
            self.args,
        ]

    def get_dependencies(self) -> List[ImageDefinition]:
        return [self.parent]

    def set_dependencies(self, deps: Iterable[ImageDefinition]) -> None:
        (self.parent,) = deps


@dataclass(eq=False)
class CopyCommandImage(StageDefinedImage):
    """Image node ending in a COPY command"""

    parent: ImageDefinition
    context: ImageDefinition
    command: str

    def local_hash_data(self, symbolic: bool):
        if symbolic or not isinstance(self.context, ContextImage):
            return self.command

        line, _ = extract_command_flags(self.command)
        try:
            tokens = split_line_tokens(line)
        except ValueError as exc:
            raise TplBuildException("Failed to prase copy command") from exc

        return [
            self.command,
            self.context.context.compute_partial_hash(  # pylint: disable=no-member
                patterns=tokens[:-1]
            ),
        ]

    def get_dependencies(self) -> List[ImageDefinition]:
        return [self.parent, self.context]

    def set_dependencies(self, deps: Iterable[ImageDefinition]) -> None:
        self.parent, self.context = deps


@dataclass(eq=False)
class SourceImage(ImageDefinition):
    """Image node representing a source image"""

    repo: str
    tag: str
    platform: str
    digest: Optional[str] = None

    def local_hash_data(self, symbolic: bool) -> Any:
        if symbolic:
            return [self.repo, self.tag, self.platform]

        if self.digest is None:
            raise ValueError("Cannot full hash SourceImage with unresolved digest")

        return self.digest


@dataclass(eq=False)
class ScratchImage(ImageDefinition):
    """Represents an empty scratch image"""

    #: The platform of the scratch image. Many build systems separate image data by
    #: platform so even though scratch images do not necessarily have platform-specific
    #: data we need to handle them each like they are different.
    platform: str

    def local_hash_data(self, symbolic: bool) -> Any:
        return self.platform


@dataclass(eq=False)
class MultiPlatformImage(StageDefinedImage):
    """
    Container image node that merges multiple other images into a single
    manifest list.
    """

    #: Mapping of platform names to images.
    images: Dict[str, ImageDefinition]

    def get_dependencies(self) -> List[ImageDefinition]:
        return list(self.images.values())

    def set_dependencies(self, deps: Iterable[ImageDefinition]) -> None:
        for platform, dep in zip(self.images, deps):
            self.images[platform] = dep

    def local_hash_data(self, symbolic: bool) -> Any:
        return list(self.images)


@dataclass(eq=False)
class BaseImage(ImageDefinition):
    """
    Image node representing a base image.

    Attributes:
    """

    #: Name of the profile this base image belongs to
    profile: str
    #: Name of the build stage
    stage: str
    #: The platform to select for this base image
    platform: str
    #: The build graph behind this base image. This can be None if the base image will
    #: not be dereferenced.
    image: Optional[ImageDefinition] = None
    #: The conent hash of the base image. Typically this is supplied from tplbuild's
    #: cached build data and is used to find the base image from an external repository.
    content_hash: Optional[str] = None
    #: The digest of the image as stored in the registry. Like :attr:`content_hash` this
    #: also typically comes from the cached build data.
    digest: Optional[str] = None

    def get_dependencies(self) -> List[ImageDefinition]:
        return [self.image] if self.image else []

    def set_dependencies(self, deps: Iterable[ImageDefinition]) -> None:
        if deps:
            (self.image,) = deps

    def local_hash_data(self, symbolic: bool) -> Any:
        if symbolic:
            return [self.profile, self.stage, self.platform]

        if self.content_hash is None:
            raise ValueError("Cannot hash BaseImage with unresolved content hash")

        return self.content_hash


@dataclass(eq=False)
class ContextImage(StageDefinedImage):
    """Image node representing a build context"""

    context: BuildContext
    platform: str

    def local_hash_data(self, symbolic: bool) -> Any:
        """
        For non-symbolic hash calculations it's actually the CopyImage nodes
        that calculate the hash based on file data in the context.
        """
        if symbolic:
            return [self.platform, self.context.symbolic_hash]
        return self.platform


@dataclass(eq=False)
class StageData:
    """
    Dataclass holding metadata about a rendered image stage.
    """

    #: The name of the build stage
    name: str
    #: The image definition
    image: ImageDefinition
    #: The stage config
    config: StageConfig
    #: If this is a base image this will be set as the appropriate base
    #: image reference.
    base_image: Optional[BaseImage] = None



import argparse

from tplbuild.cmd.common import debug_build_operations
from tplbuild.cmd.utility import CliUtility
from tplbuild.exceptions import TplBuildException
from tplbuild.tplbuild import TplBuild
from tplbuild.utils import compute_extra_vars


class BuildUtility(CliUtility):
    """CLI utility entrypoint for building top-level images"""

    def setup_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.description = "build top-level images locally"
        parser.add_argument(
            "image",
            nargs="*",
            help="Images to build. Use 'stage_name=target_name' to "
            "override the default image name for stage_name or "
            "'stage_name=' to use the stage name as the image name.",
        )
        parser.add_argument(
            "--profile",
            required=False,
            default=None,
            help="Profile to build. Defaults to default profile.",
        )
        parser.add_argument(
            "--platform",
            required=False,
            default=None,
            help="Platform to build images for. "
            "Defaults to current executor platform.",
        )
        parser.add_argument(
            "--set",
            dest="set_args",
            action="append",
            type=lambda val: (False, val),
            help="Set a custom variable for the build in the format '--set a.b=c'",
        )
        parser.add_argument(
            "--set-json",
            dest="set_args",
            action="append",
            type=lambda val: (True, val),
            help="Like --set except the value will be decoded as a JSON payload."
            " e.g. '--set-json a.b=[1,\"xyz\",null]'",
        )
        parser.add_argument(
            "--debug",
            default=False,
            const=True,
            action="store_const",
            help="Only print rendered Dockerfiles instead of building images",
        )

    async def main(self, args, tplbld: TplBuild) -> int:
        profile = args.profile or tplbld.config.default_profile
        extra_vars = compute_extra_vars(args.set_args or [])

        # Render all build stages
        stage_mapping = await tplbld.render(
            profile=profile,
            platform=args.platform,
            extra_vars=extra_vars,
        )

        # Remove push names
        for stage_data in stage_mapping.values():
            stage_data.config.push_names = []

        # Figure out what images to build, override image_names where requested.
        images_to_build = set()
        for image_arg in args.image:
            image_parts = image_arg.split("=", maxsplit=1)
            images_to_build.add(image_parts[0])
            if image_parts[0] not in stage_mapping:
                raise TplBuildException(f"Unknown build stage {repr(image_parts[0])}")
            if len(image_parts) > 1:
                stage_mapping[image_parts[0]].config.image_names = [
                    image_parts[1] or image_parts[0],
                ]

        # Only explicitly build stages that have image_names associated with them.
        # Anything else that is needed will be included implicitly in the build graph.
        stages_to_build = [
            stage
            for stage_name, stage in stage_mapping.items()
            if stage.config.image_names
            and (not images_to_build or stage_name in images_to_build)
        ]

        # Resolve the locked source image manifest content address from cached
        # build data.
        await tplbld.resolve_source_images(stages_to_build)

        # Resolve BaseImage nodes' content_hash so that their prebuilt image
        # can be referenced correctly.
        await tplbld.resolve_base_images(stages_to_build, dereference=False)

        # Create a plan of build operations to execute the requested build.
        build_ops = tplbld.plan(stages_to_build)

        # Execute the build operations.
        if args.debug:
            debug_build_operations(tplbld, build_ops)
        else:
            await tplbld.build(build_ops)

        return 0



import argparse
import asyncio
import collections
import json
import os
from typing import Dict, Set

from aioregistry import RegistryException, parse_image_name

from tplbuild.cmd.utility import CliUtility
from tplbuild.config import BuildData
from tplbuild.exceptions import TplBuildException
from tplbuild.images import BaseImage
from tplbuild.tplbuild import TplBuild


async def _run_git_command(*args):
    try:
        proc = await asyncio.create_subprocess_exec(
            *args,
            stdout=asyncio.subprocess.PIPE,
        )
        stdout, _ = await proc.communicate()
        if proc.returncode:
            raise TplBuildException("git command failed")
        return stdout.decode("utf-8")
    except OSError as exc:
        raise TplBuildException("git command failed") from exc


class BasePruneUtility(CliUtility):
    """CLI utility entrypoint for building base images"""

    def setup_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.description = (
            "Remove base images not referenced by a passed git commit. "
            "Images deleted cannot be recovered without rebuilding."
        )
        parser.add_argument(
            "git_commits",
            nargs="*",
            help="A list of git commits to preserve base images",
        )
        parser.add_argument(
            "--force",
            required=False,
            const=True,
            default=False,
            action="store_const",
            help="Do not prompt for confirmation",
        )

    def _collect_build_data(
        self, tplbld: TplBuild, build_data: BuildData, known_tags: Dict[str, Set[str]]
    ) -> None:
        """
        Add all known tags in the build data.
        """
        for profile, bases_for_profile in build_data.base.items():
            for stage, bases_for_stage in bases_for_profile.items():
                for platform, base_data in bases_for_stage.items():
                    base_image = BaseImage(
                        profile=profile,
                        stage=stage,
                        platform=platform,
                        content_hash=base_data.build_hash,
                    )
                    image_ref = parse_image_name(tplbld.get_base_image_name(base_image))
                    known_tags[image_ref.name(include_ref=False)].add(image_ref.ref)

    async def main(self, args, tplbld: TplBuild) -> int:
        """TODO"""

        known_tags: Dict[str, Set[str]] = collections.defaultdict(set)

        # Always collect the current base image build data
        self._collect_build_data(tplbld, tplbld.build_data, known_tags)

        # Collect base image build data from each git commit passed.
        if args.git_commits:
            git_path = (
                await _run_git_command(
                    "git",
                    "ls-files",
                    "--full-name",
                    os.path.join(tplbld.base_dir, ".tplbuilddata.json"),
                )
            ).rstrip()

            for commit in args.git_commits:
                build_data_json = await _run_git_command(
                    "git", "show", f"{commit}:{git_path}"
                )
                build_data = BuildData(**json.loads(build_data_json))
                self._collect_build_data(tplbld, build_data, known_tags)

        refs_to_delete = []
        for repo, repo_tags in known_tags.items():
            image_ref = parse_image_name(repo)
            tags = await tplbld.registry_client.registry_repo_tags(
                image_ref.registry, image_ref.repo
            )

            for tag in tags:
                if tag not in repo_tags:
                    refs_to_delete.append(image_ref.copy(update=dict(ref=tag)))

        if not refs_to_delete:
            print("Nothing to delete")
            return 0

        print(f"Going to delete {len(refs_to_delete)} tags")
        if not args.force:
            for ref in refs_to_delete:
                print(f"  {ref}")

            data = input("Delete tags [yN]: ")
            if data not in ("y", "Y"):
                return 1

        delete_sem = asyncio.BoundedSemaphore(8)

        async def _delete_ref(ref) -> bool:
            async with delete_sem:
                delete_okay = False
                try:
                    delete_okay = await tplbld.registry_client.ref_delete(ref)
                    if delete_okay:
                        print(f"Deleted ref {ref}")
                    else:
                        print(f"Failed to delete ref {ref}")
                except RegistryException as exc:
                    print(f"Failed to delete ref {ref}: {exc}")
            return delete_okay

        result = await asyncio.gather(*(_delete_ref(ref) for ref in refs_to_delete))
        return 0 if all(result) else 1



import platform
from typing import Tuple


def normalize_architecture(arch: str, variant: str) -> Tuple[str, str]:
    """
    Normalize the passed architecture and variant. This is copying the
    same semantics used in
    https://github.com/moby/containerd/blob/ambiguous-manifest-moby-20.10/platforms/database.go
    """
    arch, variant = arch.lower(), variant.lower()
    if arch == "i386":
        return "386", ""
    if arch in ("x86_64", "x86-64"):
        return ("amd64", "")
    if arch in ("aarch64", "arm64"):
        if variant in ("8", "v8"):
            variant = ""
        return "arm64", variant
    if arch == "armhf":
        return "arm", "v7"
    if arch == "armel":
        return "arm", "v6"
    if arch == "arm":
        if variant in ("", "7"):
            variant = "7"
        elif variant in ("5", "6", "8"):
            variant = f"v{variant}"
        return "arm", variant
    return arch, variant


def normalize_platform(os: str, arch: str, variant: str = "") -> str:
    """
    Like :meth:`normalize_architecture` except include the `os` and return
    the normalized platform string.
    """
    arch, variant = normalize_architecture(arch, variant)
    if variant:
        return f"{os}/{arch}/{variant}"
    return f"{os}/{arch}"


def split_platform(platfrm: str) -> Tuple[str, str, str]:
    """
    Split a platform string into its (os, architecture, variant) form.
    """
    parts = platfrm.split("/", maxsplit=2)
    return (
        parts[0],
        parts[1] if len(parts) > 1 else "",
        parts[2] if len(parts) > 2 else "",
    )


def normalize_platform_string(platform_str: str) -> str:
    """
    Normalize a platform string like 'linux/arm64/v8' into 'linux/arm64'.
    """
    parts = platform_str.split("/", maxsplit=2)
    if len(parts) == 1:
        return normalize_platform("linux", parts[0])
    return normalize_platform(*parts)


def client_platform(osname: str = "linux"):
    """
    Return the appropriate platform string for the client. This will ignore
    the local operating system and just use "linux" as the OS type as this
    is typically what is wanted.
    """
    return normalize_platform(osname, platform.machine(), "")



import asyncio
import logging
import os
import sys
from argparse import SUPPRESS, ArgumentParser
from typing import Callable, Dict, Mapping

import yaml
from aioregistry import (
    AsyncRegistryClient,
    ChainedCredentialStore,
    CredentialStore,
    DockerCredentialStore,
    default_credential_store,
)

from tplbuild.cmd.base_build import BaseBuildUtility
from tplbuild.cmd.base_lookup import BaseLookupUtility
from tplbuild.cmd.base_prune import BasePruneUtility
from tplbuild.cmd.build import BuildUtility
from tplbuild.cmd.publish import PublishUtility
from tplbuild.cmd.source_lookup import SourceLookupUtility
from tplbuild.cmd.source_update import SourceUpdateUtility
from tplbuild.cmd.utility import CliUtility
from tplbuild.cmd.version import VersionUtility
from tplbuild.config import UserConfig
from tplbuild.exceptions import TplBuildException
from tplbuild.tplbuild import TplBuild

LOGGER = logging.getLogger(__name__)

ALL_UTILITIES: Mapping[str, Callable[[], CliUtility]] = {
    "build": BuildUtility,
    "base-build": BaseBuildUtility,
    "base-lookup": BaseLookupUtility,
    "base-prune": BasePruneUtility,
    "publish": PublishUtility,
    "source-lookup": SourceLookupUtility,
    "source-update": SourceUpdateUtility,
    "version": VersionUtility,
}


def create_main_parser(utilities: Mapping[str, CliUtility]) -> ArgumentParser:
    """Setup the argument parser configuration for each utility."""
    parents = [
        create_base_parser(),
        create_config_parser(),
    ]

    parser = ArgumentParser(
        description="templated build tool",
        parents=parents,
    )
    subparsers = parser.add_subparsers(
        required=True,
        dest="utility",
        help="what tplbuild sub-utility to invoke",
    )

    for subcommand, utility in utilities.items():
        utility.setup_parser(subparsers.add_parser(subcommand, parents=parents))

    return parser


def create_base_parser() -> ArgumentParser:
    """
    Create shared parser for basic CLI options.
    """
    parser = ArgumentParser(description="Base tplbuild options", add_help=False)
    parser.add_argument(
        "--verbose",
        "-v",
        action="count",
        default=SUPPRESS,
    )
    parser.add_argument(
        "-C",
        "--base-dir",
        required=False,
        default=SUPPRESS,
        help="Base directory for tplbuild",
    )
    return parser


def create_config_parser() -> ArgumentParser:
    """
    Create shared parser that overrides user configuration tplbuild options.
    """
    parser = ArgumentParser(description="Use config options", add_help=False)
    parser.add_argument(
        "--user-config",
        required=False,
        default=SUPPRESS,
        help="Path to user config file",
    )
    parser.add_argument(
        "--auth-file",
        required=False,
        default=SUPPRESS,
        help="Path to the container auth file",
    )
    parser.add_argument(
        "--insecure",
        required=False,
        const=True,
        action="store_const",
        default=SUPPRESS,
        help="Disable server certificate verification",
    )
    parser.add_argument(
        "--cafile",
        required=False,
        default=SUPPRESS,
        help="SSL context CA file",
    )
    parser.add_argument(
        "--capath",
        required=False,
        default=SUPPRESS,
        help="SSL context CA directory",
    )
    parser.add_argument(
        "--load-default-certs",
        required=False,
        const=True,
        action="store_const",
        default=SUPPRESS,
        help="Load system default certs always",
    )
    parser.add_argument(
        "--build-jobs",
        required=False,
        default=SUPPRESS,
        type=int,
        help="Set max concurrent build jobs",
    )
    parser.add_argument(
        "--push-jobs",
        required=False,
        default=SUPPRESS,
        type=int,
        help="Set max concurrent push or pull jobs",
    )
    return parser


def setup_logging(verbose: int) -> None:
    """Setup tplbuild default logging based on the verbosity level"""
    internal_level, external_level = logging.WARNING, logging.CRITICAL
    if verbose > 2:
        internal_level, external_level = logging.DEBUG, logging.INFO
    elif verbose > 1:
        internal_level, external_level = logging.DEBUG, logging.WARNING
    elif verbose:
        internal_level, external_level = logging.INFO, logging.ERROR

    tplbuild_root = logging.getLogger("tplbuild")
    tplbuild_root.propagate = False
    tplbuild_root.setLevel(internal_level)

    handler = logging.StreamHandler()
    handler.setFormatter(
        logging.Formatter("%(levelname)s: %(message)s"),
    )
    tplbuild_root.addHandler(handler)

    logging.basicConfig(
        format="%(levelname)s(%(module)s): %(message)s",
        level=external_level,
    )


def load_user_config(args) -> UserConfig:
    """Load the user config. Override with settings from args as requested."""
    user_config_locations = {
        os.path.expanduser("~/.tplbuildconfig.yml"),
        os.path.join(args.base_dir, ".tplbuildconfig.yml"),
    }
    if args.user_config:
        user_config_locations = {args.user_config}
    user_config_data: Dict = {}
    for user_config_path in user_config_locations:
        try:
            with open(user_config_path, encoding="utf-8") as fconfig:
                user_config_data.update(**yaml.safe_load(fconfig))
        except IOError as exc:
            if args.user_config:
                raise TplBuildException(
                    f"Failed to read user config file: {exc}"
                ) from exc
            continue
        except (ValueError, TypeError, yaml.YAMLError) as exc:
            raise TplBuildException(f"Failed to load user config: {exc}") from exc
    try:
        user_config = UserConfig(**user_config_data)
    except ValueError as exc:
        raise TplBuildException(f"Failed to load user config: {exc}") from exc

    if args.auth_file:
        user_config.auth_file = args.auth_file
    if args.insecure:
        user_config.ssl_context.insecure = True
    if args.cafile:
        user_config.ssl_context.cafile = args.cafile
    if args.capath:
        user_config.ssl_context.capath = args.capath
    if args.load_default_certs:
        user_config.ssl_context.load_default_certs = True
    if args.build_jobs is not None:
        if args.build_jobs <= 0:
            user_config.build_jobs = os.cpu_count() or 4
        else:
            user_config.build_jobs = args.build_jobs
    if args.push_jobs is not None:
        if args.push_jobs <= 0:
            user_config.push_jobs = os.cpu_count() or 4
        else:
            user_config.push_jobs = args.push_jobs
    return user_config


def create_registry_client(user_config: UserConfig) -> AsyncRegistryClient:
    """Create an AsyncRegistryClient context from the passed arguments."""
    creds: CredentialStore = default_credential_store()
    if user_config.auth_file:
        try:
            creds = ChainedCredentialStore(
                DockerCredentialStore.from_file(user_config.auth_file),
                creds,
            )
        except FileNotFoundError as exc:
            raise TplBuildException(
                f"could not open auth file {repr(user_config.auth_file)}"
            ) from exc

    return AsyncRegistryClient(
        creds=creds,
        ssl_context=user_config.ssl_context.create_context(),
    )


def create_tplbld(
    args, user_config: UserConfig, registry_client: AsyncRegistryClient
) -> TplBuild:
    """Create a TplBuild context from the passed arguments."""
    return TplBuild.from_path(
        args.base_dir, user_config=user_config, registry_client=registry_client
    )


def apply_default_args(args) -> None:
    """
    Apply default valeus to CLI arguments as needed. The normal default behavior
    of argparse does not work well with parsers shared across subparsers.
    """
    defaults = dict(
        verbose=0,
        base_dir=".",
        user_config=None,
        auth_file=None,
        insecure=False,
        cafile=None,
        capath=None,
        load_default_certs=False,
        build_jobs=None,
        push_jobs=None,
    )
    for key, val in defaults.items():
        setattr(args, key, getattr(args, key, val))


async def amain() -> int:
    """Parse CLI options, setup logging, then invoke the requested utility"""
    utilities = {
        subcommand: utility_cls() for subcommand, utility_cls in ALL_UTILITIES.items()
    }

    parser = create_main_parser(utilities)
    args = parser.parse_args()
    apply_default_args(args)
    setup_logging(args.verbose)

    try:
        user_config = load_user_config(args)
        async with create_registry_client(user_config) as registry_client:
            async with create_tplbld(args, user_config, registry_client) as tplbld:
                return await utilities[args.utility].main(args, tplbld)
    except TplBuildException as exc:
        sys.stderr.write(f"{exc}\n")
        if exc.more_message:
            sys.stderr.write(f"{exc.more_message}\n")
        LOGGER.debug("got top level tplbuild exception", exc_info=True)
        return 1
    except Exception:  # pylint: disable=broad-except
        LOGGER.exception("Unexpected top-level exception")
        return 2


def main() -> int:
    """Synchronous entry point"""
    return asyncio.run(amain())


if __name__ == "__main__":
    sys.exit(main())



import argparse

from tplbuild._version import __version__ as version
from tplbuild.cmd.utility import CliUtility
from tplbuild.tplbuild import TplBuild


class VersionUtility(CliUtility):
    """CLI utility entrypoint for building base images"""

    def setup_parser(self, parser: argparse.ArgumentParser) -> None:
        parser.description = "Print tplbuild version information"

    async def main(self, args, tplbld: TplBuild) -> int:
        """Just print the current tplbuild version"""

        print(version)
        return 0



import contextlib
import json
import os
import tempfile
from typing import Any, Dict, Iterable, List, Sequence, Tuple


def line_reader(document: str) -> Iterable[Tuple[int, int, str]]:
    """
    Yield lines from `document` in the form (source_character_position,
    source_line_position, line_content).

    Lines will have leading and trailing whitespace
    stripped. Lines that being with a '#' character will be omitted. Lines that
    end with a single backslash character will be treated as continuations with
    the following line concatenated onto itself, not including the backslash or
    line feed character.

    `source_character_position` will give the character index that the line
    begins in the original passed `document`.

    `source_line_position` will give the line index that the line begins in
    the original passed `document`. This may differ than the output line index
    in the precense of empty lines, comments, and line continuations.
    """
    lines = []
    line_start_idx = 0
    line_last_ch = ""
    for idx, ch in enumerate(document):
        if ch == "\n":
            end_idx = idx - 1 if line_last_ch == "\r" else idx
            lines.append((line_start_idx, document[line_start_idx:end_idx]))
            line_start_idx = idx + 1
        elif line_last_ch == "\r":
            lines.append((line_start_idx, document[line_start_idx : idx - 1]))
            line_start_idx = idx
        line_last_ch = ch
    if line_start_idx < len(document):
        lines.append((line_start_idx, document[line_start_idx:]))

    idx = -1
    start_line_pos = -1
    start_line_idx = -1
    line_parts: List[str] = []
    for idx, (line_pos, line_part) in enumerate(lines):
        line_part = line_part.rstrip()
        if not line_parts and not line_part:
            continue
        if not line_parts:
            start_line_pos = line_pos
            start_line_idx = idx
        if line_part.lstrip().startswith("#"):
            continue
        if line_part.endswith("\\") and not line_part.endswith("\\\\"):
            line_parts.append(line_part[:-1])
            continue

        line = ("".join(line_parts) + line_part).strip()
        line_parts.clear()
        if line:
            yield start_line_pos, start_line_idx, line

    line = "".join(line_parts).strip()
    if line:
        yield start_line_pos, start_line_idx, line


@contextlib.contextmanager
def open_and_swap(filename, mode="w+b", buffering=-1, encoding=None, newline=None):
    """
    Open a file for writing and relink it to the desired path in an atomic
    operation when the file is closed without an exception. This prevents
    the existing file data from being lost if an unexpected failure occurs
    while writing the file.
    """
    fd, tmppath = tempfile.mkstemp(
        dir=os.path.dirname(filename) or ".",
        text="b" not in mode,
    )
    fh = None
    try:
        fh = open(
            fd,
            mode=mode,
            buffering=buffering,
            encoding=encoding,
            newline=newline,
            closefd=False,
        )
        yield fh
        os.rename(tmppath, filename)
        tmppath = None
    finally:
        if fh is not None:
            fh.close()
        os.close(fd)
        if tmppath is not None:
            os.unlink(tmppath)


def split_line_tokens(line: str) -> List[str]:
    """
    Split a Dockerfile command into tokens. For commands that can accept flags
    this should be called with the `line` portion returned from
    `extract_command_flags`
    """
    line = line.strip()
    if line.startswith("["):
        parts = json.loads(line)
        if not isinstance(parts, list):
            raise ValueError("Expected JSON list")
        if not all(isinstance(part, str) for part in parts):
            raise ValueError("Expected list of strings")
        return parts
    return line.split()


def extract_command_flags(line: str) -> Tuple[str, Dict[str, str]]:
    """
    Some Dockerfile commands support flags in the form "--name=value" at the start of
    the command. This extracts the commands with those flags removed and returns a
    mapping of the extracted flags.

    Notes:
        As far as I know there is no escaping layer here. Names that contain spaces
        or equal signs cannot be represented as flag names or values.

    Returns: (line, flags) tuple
        line: The line with the flags removed.
        flags: The mapping of flag names to flag values that was extracted
    """

    def _skip_ws(pos: int) -> int:
        while pos < len(line) and line[pos].isspace():
            pos += 1
        return pos

    pos = _skip_ws(0)
    flags = {}
    while pos + 1 < len(line) and line[pos : pos + 2] == "--":
        space = line.find(" ", pos + 2)
        if space == -1:
            space = len(line)

        parts = line[pos + 2 : space].split("=", 1)
        if len(parts) == 2:
            flags[parts[0]] = parts[1]
        else:
            flags[parts[0]] = ""
        pos = _skip_ws(space)

    if not flags:
        return line, {}
    return line[pos:], flags


def format_command_with_flags(line: str, flags: Dict[str, str]) -> str:
    """
    The inverse of _extract_command_flags

    Returns:
        The command line with the flags added into the front of the command.
    """
    if not flags:
        return line
    flags_str = " ".join(f"--{key}={val}" for key, val in flags.items())
    if not line:
        return flags_str
    return f"{flags_str} {line}"


def ignore_escape(path: str) -> str:
    """
    Escape a path appropriately for a docker ignore file.
    """
    special_chars = "\\*?[]"
    return "".join("\\" + ch if ch in special_chars else ch for ch in path)


def deep_merge_json(lhs: Any, rhs: Any) -> Any:
    """
    Merge JSON-like data `rhs` into `lhs`. This will only merge dicts, lists
    will simply replace rather than concatenate.

    This will not modify any objects referenced by `lhs` or `rhs` but may
    return an object which points to objects originally in `lhs` and `rhs`.
    """

    if not isinstance(lhs, dict) or not isinstance(rhs, dict):
        return rhs
    if not rhs:
        return lhs
    if not lhs:
        return rhs

    result = dict(lhs)
    for key, val in rhs.items():
        lval = result.setdefault(key, val)
        if lval is not val:
            result[key] = deep_merge_json(lval, val)

    return result


def compute_extra_vars(set_args: Sequence[Tuple[bool, str]]) -> Dict[str, Any]:
    """
    Convert arguments passed to --set and --set-json into a dictionary of
    arguments to pass to the render. Later arguments have precedence over
    earlier arguments.
    """
    result: Dict[str, Any] = {}

    def _set_value(key: str, val: Any):
        obj = result
        key_parts = key.split(".")
        for key_part in key_parts[:-1]:
            obj = obj.setdefault(key_part, {})
            if not isinstance(obj, dict):
                obj[key_part] = {}
        obj[key_parts[-1]] = val

    for is_json, set_arg in set_args:
        arg_type = "--set-json" if is_json else "--set"

        equal_pos = set_arg.find("=")
        if equal_pos == -1:
            raise ValueError(f"Invalid {arg_type} value, expected '='")

        key = set_arg[:equal_pos]
        val = set_arg[equal_pos + 1 :]
        if is_json:
            try:
                val = json.loads(val)
            except ValueError as exc:
                raise ValueError(f"Invalid {arg_type} value JSON") from exc

        key_parts = key.split(".")
        merge_value = val
        for key_part in reversed(key_parts):
            if not key_part:
                raise ValueError(f"Cannot {arg_type} variable with empty key part")
            merge_value = {key_part: merge_value}  # type: ignore

        result = deep_merge_json(result, merge_value)

    return result
